mod commands;
mod markups;
mod state;
mod utils;

use sea_orm::prelude::*;
use std::sync::Arc;
use teloxide::{
    adaptors::{throttle::Limits, Throttle},
    dispatching::dialogue::InMemStorage,
    filter_command,
    prelude::*,
    utils::command::BotCommands,
};
use tokio::sync::Mutex;

type MainDialogue = Dialogue<State, InMemStorage<State>>;
type PackagedHandler<T> = Arc<
    Mutex<
        Option<
            Box<
                dyn FnOnce(
                        Throttle<Bot>,
                        Message,
                        DatabaseConnection,
                        MainDialogue,
                        T,
                    ) -> crate::PinnedFuture<'static, crate::Result<()>>
                    + Send
                    + Sync,
            >,
        >,
    >,
>;

#[derive(BotCommands, Clone, Copy)]
#[command(rename_rule = "snake_case")]
enum Command {
    #[command()]
    Help,
    #[command()]
    AddAccount,
    #[command()]
    GetAccount,
    #[command()]
    GetAccounts,
    #[command()]
    SetMasterPass,
    #[command()]
    Delete,
    #[command()]
    DeleteAll,
    #[command()]
    Export,
}

#[derive(Default, Clone)]
pub enum State {
    #[default]
    Start,
    GetAccountName(PackagedHandler<String>),
    GetMasterPass(PackagedHandler<String>),
    GetLogin(PackagedHandler<String>),
    GetPassword(PackagedHandler<String>),
}

pub fn get_dispatcher(
    token: String,
    db: DatabaseConnection,
) -> Dispatcher<Throttle<Bot>, crate::Error, teloxide::dispatching::DefaultKey> {
    use dptree::{case, deps, endpoint};
    let bot = Bot::new(token).throttle(Limits::default());

    let command_handler = filter_command::<Command, _>()
        .branch(case![Command::Help].endpoint(commands::help))
        .branch(case![Command::AddAccount].endpoint(commands::add_account))
        .branch(case![Command::GetAccount].endpoint(commands::get_account))
        .branch(case![Command::GetAccounts].endpoint(commands::get_accounts))
        .branch(case![Command::SetMasterPass].endpoint(commands::set_master_pass))
        .branch(case![Command::Delete].endpoint(commands::delete))
        .branch(case![Command::DeleteAll].endpoint(commands::delete_all))
        .branch(case![Command::Export].endpoint(commands::export));

    let message_handler = Update::filter_message()
        .map_async(utils::delete_message)
        .enter_dialogue::<Update, InMemStorage<State>, State>()
        .branch(case![State::GetAccountName(next)].endpoint(state::get_account_name))
        .branch(case![State::GetMasterPass(next)].endpoint(state::get_master_pass))
        .branch(case![State::GetLogin(next)].endpoint(state::get_login))
        .branch(case![State::GetPassword(next)].endpoint(state::get_password))
        .branch(command_handler)
        .branch(endpoint(commands::default));

    Dispatcher::builder(bot, message_handler)
        .dependencies(deps![db, InMemStorage::<State>::new()])
        .enable_ctrlc_handler()
        .build()
}
